@@ -23,6 +23,7 @@ import collections.abc as cabc
 import contextlib
 import ctypes
 import datetime
+from distutils.version import LooseVersion
 import functools
 import glob
 import itertools
@@ -38,6 +39,7 @@ import operator
 
 # adding imports from further xonsh modules is discouraged to avoid circular
 # dependencies
+from xonsh import __version__
 from xonsh.lazyasd import LazyObject, LazyDict, lazyobject
 from xonsh.platform import (has_prompt_toolkit, scandir, DEFAULT_ENCODING,
                             ON_LINUX, ON_WINDOWS, PYTHON_VERSION_INFO,
@@ -204,6 +206,28 @@ class EnvPath(collections.MutableSequence):
             return False
         return all(map(operator.eq, self, other))
 
+    def _repr_pretty_(self, p, cycle):
+        """ Pretty print path list """
+        if cycle:
+            p.text('EnvPath(...)')
+        else:
+            with p.group(1, 'EnvPath(\n[', ']\n)'):
+                for idx, item in enumerate(self):
+                    if idx:
+                        p.text(',')
+                        p.breakable()
+                    p.pretty(item)
+
+    def __add__(self, other):
+        if isinstance(other, EnvPath):
+            other = other._l
+        return EnvPath(self._l + other)
+
+    def __radd__(self, other):
+        if isinstance(other, EnvPath):
+            other = other._l
+        return EnvPath(other + self._l)
+
 
 class DefaultNotGivenType(object):
     """Singleton for representing when no default value is given."""
@@ -238,6 +262,25 @@ def _is_not_lparen_and_rparen(lparens, rtok):
     return rtok.type == 'RPAREN' and any(x != 'LPAREN' for x in lparens)
 
 
+def balanced_parens(line, mincol=0, maxcol=None, lexer=None):
+    """Determines if parentheses are balanced in an expression."""
+    line = line[mincol:maxcol]
+    if lexer is None:
+        lexer = builtins.__xonsh_execer__.parser.lexer
+    if '(' not in line and ')' not in line:
+        return True
+    cnt = 0
+    lexer.input(line)
+    for tok in lexer:
+        if tok.type in LPARENS:
+            cnt += 1
+        elif tok.type == 'RPAREN':
+            cnt -= 1
+        elif tok.type == 'ERRORTOKEN' and ')' in tok.value:
+            cnt -= 1
+    return cnt == 0
+
+
 def find_next_break(line, mincol=0, lexer=None):
     """Returns the column number of the next logical break in subproc mode.
     This function may be useful in finding the maxcol argument of
@@ -272,7 +315,7 @@ def find_next_break(line, mincol=0, lexer=None):
 
 def subproc_toks(line, mincol=-1, maxcol=None, lexer=None, returnline=False,
                  greedy=False):
-    """Excapsulates tokens in a source code line in a uncaptured
+    """Encapsulates tokens in a source code line in a uncaptured
     subprocess ![] starting at a minimum column. If there are no tokens
     (ie in a comment line) this returns None. If greedy is True, it will encapsulate
     normal parentheses. Greedy is False by default.
@@ -318,7 +361,9 @@ def subproc_toks(line, mincol=-1, maxcol=None, lexer=None, returnline=False,
         if pos < mincol:
             continue
         toks.append(tok)
-        if tok.type == 'NEWLINE':
+        if tok.type == 'WS' and tok.value == '\\':
+            pass  # line continuation
+        elif tok.type == 'NEWLINE':
             break
         elif tok.type == 'DEDENT':
             # fake a newline when dedenting without a newline
@@ -403,13 +448,13 @@ def _have_open_triple_quotes(s):
     return open_triple
 
 
-@lazyobject
-def LINE_CONTINUATION():
-    """ The line contiuation characters used in subproc mode. In interactive
-         mode on Windows the backslash must be preseeded by a space. This is because
-         paths on windows may end in a backspace.
+def get_line_continuation():
+    """ The line continuation characters used in subproc mode. In interactive
+         mode on Windows the backslash must be preceded by a space. This is because
+         paths on Windows may end in a backslash.
     """
-    if ON_WINDOWS and builtins.__xonsh_env__.get('XONSH_INTERACTIVE'):
+    if (ON_WINDOWS and hasattr(builtins, '__xonsh_env__') and
+            builtins.__xonsh_env__.get('XONSH_INTERACTIVE', False)):
         return ' \\'
     else:
         return '\\'
@@ -423,10 +468,13 @@ def get_logical_line(lines, idx):
     """
     n = 1
     nlines = len(lines)
+    linecont = get_line_continuation()
+    while idx > 0 and lines[idx-1].endswith(linecont):
+        idx -= 1
+    start = idx
     line = lines[idx]
-    linecont = str(LINE_CONTINUATION)
     open_triple = _have_open_triple_quotes(line)
-    while (line.endswith(linecont) or open_triple) and idx < nlines:
+    while (line.endswith(linecont) or open_triple) and idx < nlines - 1:
         n += 1
         idx += 1
         if line.endswith(linecont):
@@ -434,13 +482,14 @@ def get_logical_line(lines, idx):
         else:
             line = line + '\n' + lines[idx]
         open_triple = _have_open_triple_quotes(line)
-    return line, n
+    return line, n, start
 
 
 def replace_logical_line(lines, logical, idx, n):
     """Replaces lines at idx that may end in line continuation with a logical
     line that spans n lines.
     """
+    linecont = get_line_continuation()
     if n == 1:
         lines[idx] = logical
         return
@@ -454,7 +503,7 @@ def replace_logical_line(lines, logical, idx, n):
             logical = ''
         else:
             # found space to split on
-            lines[i] = logical[:b] + str(LINE_CONTINUATION)
+            lines[i] = logical[:b] + linecont
             logical = logical[b:]
     lines[idx+n-1] = logical
 
@@ -486,7 +535,7 @@ def subexpr_from_unbalanced(expr, ltok, rtok):
 
 
 def subexpr_before_unbalanced(expr, ltok, rtok):
-    """Obtains the expression prior to last unblanced left token."""
+    """Obtains the expression prior to last unbalanced left token."""
     subexpr, _, post = expr.rpartition(ltok)
     nrtoks_in_post = post.count(rtok)
     while nrtoks_in_post != 0:
@@ -857,12 +906,11 @@ def suggestion_sort_helper(x, y):
 
 def escape_windows_cmd_string(s):
     """Returns a string that is usable by the Windows cmd.exe.
-    The escaping is based on details here and emperical testing:
+    The escaping is based on details here and empirical testing:
     http://www.robvanderwoude.com/escapechars.php
     """
-    for c in '()%!^<>&|"':
+    for c in '^()%!<>&|"':
         s = s.replace(c, '^' + c)
-    s = s.replace('/?', '/.')
     return s
 
 
@@ -1409,7 +1457,7 @@ def is_history_backend(x):
 
 def is_dynamic_cwd_width(x):
     """ Determine if the input is a valid input for the DYNAMIC_CWD_WIDTH
-    environement variable.
+    environment variable.
     """
     return (isinstance(x, tuple) and
             len(x) == 2 and
@@ -1445,7 +1493,7 @@ RE_HISTORY_TUPLE = LazyObject(
 
 
 def to_history_tuple(x):
-    """Converts to a canonincal history tuple."""
+    """Converts to a canonical history tuple."""
     if not isinstance(x, (cabc.Sequence, float, int)):
         raise ValueError('history size must be given as a sequence or number')
     if isinstance(x, str):
@@ -1594,7 +1642,7 @@ def format_std_prepost(template, env=None):
     except Exception:
         print_exception()
     # \001\002 is there to fool pygments into not returning an empty string
-    # for potentially empty input. This happend when the template is just a
+    # for potentially empty input. This happens when the template is just a
     # color code with no visible text.
     invis = '\001\002'
     s = shell.format_color(invis + s + invis, force_string=True)
@@ -1602,7 +1650,7 @@ def format_std_prepost(template, env=None):
     return s
 
 
-_RE_STRING_START = "[bBprRuU]*"
+_RE_STRING_START = "[bBprRuUf]*"
 _RE_STRING_TRIPLE_DOUBLE = '"""'
 _RE_STRING_TRIPLE_SINGLE = "'''"
 _RE_STRING_DOUBLE = '"'
@@ -1707,19 +1755,13 @@ def check_for_partial_string(x):
         return (string_indices[-2], string_indices[-1], starting_quote[-1])
 
 
-# regular expressions for matching enviroment variables
+# regular expressions for matching environment variables
 # i.e $FOO, ${'FOO'}
 @lazyobject
 def POSIX_ENVVAR_REGEX():
     pat = r"""\$({(?P<quote>['"])|)(?P<envvar>\w+)((?P=quote)}|(?:\1\b))"""
     return re.compile(pat)
 
-if ON_WINDOWS:
-    # i.e %FOO%
-    @lazyobject
-    def WINDOWS_ENVVAR_REGEX():
-        return re.compile(r"%(?P<envvar>\w+)%")
-
 
 def expandvars(path):
     """Expand shell variables of the forms $var, ${var} and %var%.
@@ -1731,13 +1773,6 @@ def expandvars(path):
     elif isinstance(path, pathlib.Path):
         # get the path's string representation
         path = str(path)
-    if ON_WINDOWS and '%' in path:
-        for match in WINDOWS_ENVVAR_REGEX.finditer(path):
-            name = match.group('envvar')
-            if name in env:
-                ensurer = env.get_ensurer(name)
-                value = ensurer.detype(env[name])
-                path = WINDOWS_ENVVAR_REGEX.sub(value, path, count=1)
     if '$' in path:
         for match in POSIX_ENVVAR_REGEX.finditer(path):
             name = match.group('envvar')
@@ -1766,7 +1801,7 @@ def backup_file(fname):
 
 
 def normabspath(p):
-    """Retuns as normalized absolute path, namely, normcase(abspath(p))"""
+    """Returns as normalized absolute path, namely, normcase(abspath(p))"""
     return os.path.normcase(os.path.abspath(p))
 
 
@@ -1853,7 +1888,11 @@ def _iglobpath(s, ignore_case=False, sort_result=None):
 
 def iglobpath(s, ignore_case=False, sort_result=None):
     """Simple wrapper around iglob that also expands home and env vars."""
-    return _iglobpath(s, ignore_case=ignore_case, sort_result=sort_result)[0]
+    try:
+        return _iglobpath(s, ignore_case=ignore_case, sort_result=sort_result)[0]
+    except IndexError:
+        # something went wrong in the actual iglob() call
+        return iter(())
 
 
 def ensure_timestamp(t, datetime_format=None):
@@ -1935,3 +1974,72 @@ def uncapturable(f):
 def carriage_return():
     """Writes a carriage return to stdout, and nothing else."""
     print('\r', flush=True, end='')
+
+
+def deprecated(deprecated_in=None, removed_in=None):
+    """Parametrized decorator that deprecates a function in a graceful manner.
+
+    Updates the decorated function's docstring to mention the version
+    that deprecation occurred in and the version it will be removed
+    in if both of these values are passed.
+
+    When removed_in is not a release equal to or less than the current
+    release, call ``warnings.warn`` with details, while raising
+    ``DeprecationWarning``.
+
+    When removed_in is a release equal to or less than the current release,
+    raise an ``AssertionError``.
+
+    Parameters
+    ----------
+    deprecated_in : str
+        The version number that deprecated this function.
+    removed_in : str
+        The version number that this function will be removed in.
+    """
+    message_suffix = _deprecated_message_suffix(deprecated_in, removed_in)
+    if not message_suffix:
+        message_suffix = ''
+
+    def decorated(func):
+        warning_message = '{} has been deprecated'.format(func.__name__)
+        warning_message += message_suffix
+
+        @functools.wraps(func)
+        def wrapped(*args, **kwargs):
+            _deprecated_error_on_expiration(func.__name__, removed_in)
+            func(*args, **kwargs)
+            warnings.warn(warning_message, DeprecationWarning)
+
+        wrapped.__doc__ = (
+            '{}\n\n{}'.format(wrapped.__doc__, warning_message)
+            if wrapped.__doc__ else warning_message)
+
+        return wrapped
+    return decorated
+
+
+def _deprecated_message_suffix(deprecated_in, removed_in):
+    if deprecated_in and removed_in:
+        message_suffix = (
+            ' in version {} and will be removed in version {}'.format(
+                deprecated_in, removed_in))
+    elif deprecated_in and not removed_in:
+        message_suffix = (
+            ' in version {}'.format(deprecated_in))
+    elif not deprecated_in and removed_in:
+        message_suffix = (
+            ' and will be removed in version {}'.format(removed_in))
+    else:
+        message_suffix = None
+
+    return message_suffix
+
+
+def _deprecated_error_on_expiration(name, removed_in):
+    if not removed_in:
+        return
+    elif LooseVersion(__version__) >= LooseVersion(removed_in):
+        raise AssertionError(
+            '{} has passed its version {} expiry date!'.format(
+                name, removed_in))
